// $Id: Mp3Player.java,v 1.14 2005/02/22 22:07:41 xi Exp $

/******************************************************************************
*    MusicControl - An advanced music player
*    Copyright (C) 2002-2005  Christian Laursen <xi@borderworlds.dk>
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

package dk.borderworlds;

import java.io.*;
import java.util.regex.*;

public class Mp3Player extends MusicPlayer implements Runnable {
	private Song song;
	private Process proc;
	private BufferedInputStream bis;
	String name;
	private Thread inputreader;
	private String status;
	private byte[] statusLine;
	private String title, filename;
	private int pausePosition;
	private MusicConfiguration configuration;
	private MusicPipe pipe;

	public Mp3Player (MusicConfiguration configuration) {
		song = null;
		proc = null;
		this.configuration = configuration;
	}

	public void playSong (Song song) {
		this.song = song;
		playSong(0);
	}

	public void playSong (int offset) {
		if (offset == 0) {
			status = "Starting";
		} else {
			status = "Resuming";
		}
		title = song.getFilename();
		filename = song.getFilename();
		Runtime rt = Runtime.getRuntime();
		if (proc != null) {
			proc.destroy();
			while (bis != null) {
				try {
					Thread.sleep(300);
				} catch (Exception e) {
				}
			}
			try {
				Thread.sleep(300);
			} catch (Exception e) {
			}
			proc = null;
		}
		try {
			String device = configuration.getDevice();
			int args = (pausePosition == 0) ? 3 : 5;
			if (device.equals("STDOUT")) {
				args += 1;
			}
			String[] cmdarray = new String[args];
			cmdarray[0] = "mpg123";
			cmdarray[1] = "-v";
			if (device.equals("STDOUT")) {
				cmdarray[2] = "-s";
			}
			if (pausePosition != 0) {
				cmdarray[args - 3] = "-k";
				cmdarray[args - 2] = "" + pausePosition;
			}
			cmdarray[args - 1] = song.getDirectory() + File.separator + song.getFilename();
			proc = rt.exec(cmdarray);
			InputStream is = proc.getErrorStream();
			pipe = new MusicPipe(proc.getInputStream());
			bis = new BufferedInputStream(is);
			inputreader = new Thread (this);
			inputreader.start();
			status = "Playing";
		} catch (Exception e) {
		}
	}

	public void pauseSong () {
		if (statusLine != null) {
			int length = 0;
			while ((statusLine.length > 16 + length) && (statusLine[16 + length] != 32) && (length < 6)) {
				++length;
			}
			String frame = new String(statusLine, 16, length);
			pausePosition = Integer.parseInt(frame);
		} else {
			pausePosition = 0;
		}

		if (proc != null) {
			proc.destroy();
			while (bis != null) {
				try {
					Thread.sleep(300);
				} catch (Exception e) {
				}
			}
			proc = null;
		}
	}

	public void resumeSong () {
		if (pausePosition != 0) {
			playSong(pausePosition);
			pausePosition = 0;
		}
	}

	public void stopSong () {
		if (proc != null) {
			proc.destroy();
			while (bis != null) {
				try {
					Thread.sleep(300);
				} catch (Exception e) {
				}
			}
			proc = null;
			song = null;
		}
		try {
			Thread.sleep(300);
		} catch (Exception e) {
		}
	}

	public String getStatus () {
		if (status.equals("Retrying")) {
			playSong(0);
			return "Retrying";
		}
		if (status.equals("Playing") && bis == null) {
			if (pausePosition == 0) {
				status = "Stopped";
				song = null;
			} else {
				status = "Paused";
			}
		}
		return status;
	}

	public String getFilename () {
		if (song == null) {
			return "(none)";
		} else {
			return filename;
		}
	}

	public String getTitle () {
		if (song == null) {
			return "(none)";
		} else {
			return title;
		}
	}

	public String getPosition () {
		if (statusLine != null) {
			return new String(statusLine, 0, 5);
		} else {
			return("N/A");
		}
	}

	public String getLength () {
		if ((statusLine != null) && (statusLine[0] >= 48) && (statusLine[0] <= 57)) {
			int length = (Integer.parseInt(new String(statusLine, 0,2)) * 100 * 60 +
						  Integer.parseInt(new String(statusLine, 8,2)) * 100 * 60 +
						  Integer.parseInt(new String(statusLine, 3,2)) * 100 +
						  Integer.parseInt(new String(statusLine, 11,2)) * 100 +
						  Integer.parseInt(new String(statusLine, 6,2)) +
						  Integer.parseInt(new String(statusLine, 14,2))) / 100;
			return ((length / 60 < 10) ? "0" : "") + (length / 60) + ":" +
				((length % 60 < 10) ? "0" : "") + (length % 60);
		} else {
			return("N/A");
		}
	}

	public boolean acceptsType (String type) {
		return type.equals("mp3");
	}

	public Song getSong () {
		return song;
	}

	public void run () {
		byte[] b, lsl;
		int i, j, k, lsll;
		boolean cont = true;

		b = new byte[1024];
		statusLine = new byte[22];
		lsl = new byte[256];
		lsll = 0;

		for (i = 0; i < 22; ++i) {
			statusLine[i] = 32;
		}

		Pattern timePattern = Pattern.compile("^Frame#\\s+(\\d+).*Time:\\s+(\\d{2}:\\d{2}\\.\\d{2})\\s+\\[(\\d{2}:\\d{2}\\.\\d{2})\\]");
		Pattern titlePattern = Pattern.compile("^Title\\s+:\\s+(.*?)\\s+Artist:");
		Pattern errorPattern = Pattern.compile("^Can't");
		Matcher m;

		while (cont) {
			try {
				i = bis.read(b, 0, 1024);
				if (i > 0) {
					for (j = 0; j < i; ++j) {
						if (b[j] == 10 || b[j] == 13) {
							String line = new String(lsl, 0, lsll);
							m = timePattern.matcher(line);
							if (m.find()) {
								statusLine = (m.group(2) + m.group(3) + m.group(1)).getBytes();
							}

							m = titlePattern.matcher(line);
							if (m.find()) {
								title = m.group(1);
							}

							m = errorPattern.matcher(line);
							if (m.find()) {
								status = "Retrying";
							}

							lsll = 0;
						} else {
							lsl[lsll++] = b[j];
						}
					}
				} else if (i == -1) {
					bis.close();
					bis = null;
					cont = false;
				}
			} catch (Exception e) {
				if (e.getClass().getName().equals("java.io.IOException")) {
					bis = null;
					cont = false;
				}
			}
		}
	}
}
